import PropTypes from 'prop-types';
import { Children, Component, cloneElement, findDOMNode, findVDom, h, isValidElement, options, render } from 'zreact';

var version = "15.1.0"; // trick libraries to think we are react
var ELEMENTS = "a abbr address area article aside audio b base bdi bdo big blockquote body br button canvas caption cite code col colgroup data datalist dd del details dfn dialog div dl dt em embed fieldset figcaption figure footer form h1 h2 h3 h4 h5 h6 head header hgroup hr html i iframe img input ins kbd keygen label legend li link main map mark menu menuitem meta meter nav noscript object ol optgroup option output p param picture pre progress q rp rt ruby s samp script section select small source span strong style sub summary sup table tbody td textarea tfoot th thead time title tr track u ul var video wbr circle clipPath defs ellipse g image line linearGradient mask path pattern polygon polyline radialGradient rect stop svg text tspan".split(" ");
var REACT_ELEMENT_TYPE = (typeof Symbol !== "undefined" && Symbol.for && Symbol.for("react.element")) || 0xeac7;
var COMPONENT_WRAPPER_KEY = typeof Symbol !== "undefined" ? Symbol.for("__zreactCompatWrapper") : "__zreactCompatWrapper";
// don"t autobind these methods since they already have guaranteed context.
var AUTOBIND_BLACKLIST = {
    componentDidMount: 1,
    componentDidUnmount: 1,
    componentDidUpdate: 1,
    componentWillMount: 1,
    componentWillReceiveProps: 1,
    componentWillUnmount: 1,
    componentWillUpdate: 1,
    constructor: 1,
    render: 1,
    shouldComponentUpdate: 1,
};
var CAMEL_PROPS = /^(?:accent|alignment|arabic|baseline|cap|clip|color|fill|flood|font|glyph|horiz|marker|overline|paint|stop|strikethrough|stroke|text|underline|unicode|units|v|vector|vert|word|writing|x)[A-Z]/;
var BYPASS_HOOK = {};
/*global process*/
var DEV = typeof process === "undefined" || !process.env || process.env.NODE_ENV !== "production";
// a component that renders nothing. Used to replace components for unmountComponentAtNode.
function EmptyComponent() { return null; }
// make react think we"re react.
var VNode = h("a", null).constructor;
VNode.prototype.$$typeof = REACT_ELEMENT_TYPE;
VNode.prototype.zreactCompatUpgraded = false;
VNode.prototype.zreactCompatNormalized = false;
Object.defineProperty(VNode.prototype, "type", {
    get: function () { return this.nodeName; },
    set: function (v) { this.nodeName = v; },
    configurable: true,
});
Object.defineProperty(VNode.prototype, "props", {
    get: function () { return this.attributes; },
    set: function (v) { this.attributes = v; },
    configurable: true,
});
var oldEventHook = options.event;
options.event = function (e) {
    if (oldEventHook) {
        e = oldEventHook(e);
    }
    e.persist = Object;
    e.nativeEvent = e;
    return e;
};
var oldVnodeHook = options.vnode;
options.vnode = function (vnode) {
    if (!vnode.zreactCompatUpgraded) {
        vnode.zreactCompatUpgraded = true;
        var tag = vnode.nodeName;
        var attrs = vnode.attributes = extend({}, vnode.attributes);
        if (typeof tag === "function") {
            if (tag[COMPONENT_WRAPPER_KEY] === true || (tag.prototype && "isReactComponent" in tag.prototype)) {
                if (vnode.children && String(vnode.children) === "") {
                    vnode.children = undefined;
                }
                if (vnode.children) {
                    attrs.children = vnode.children;
                }
                if (!vnode.zreactCompatNormalized) {
                    normalizeVNode(vnode);
                }
                handleComponentVNode(vnode);
            }
        }
        else {
            if (vnode.children && String(vnode.children) === "") {
                vnode.children = undefined;
            }
            if (vnode.children) {
                attrs.children = vnode.children;
            }
            if (attrs.defaultValue) {
                if (!attrs.value && attrs.value !== 0) {
                    attrs.value = attrs.defaultValue;
                }
                delete attrs.defaultValue;
            }
            handleElementVNode(vnode, attrs);
        }
    }
    if (oldVnodeHook) {
        oldVnodeHook(vnode);
    }
};
function handleComponentVNode(vnode) {
    var tag = vnode.nodeName;
    var a = vnode.attributes;
    vnode.attributes = {};
    if (tag.defaultProps) {
        extend(vnode.attributes, tag.defaultProps);
    }
    if (a) {
        extend(vnode.attributes, a);
    }
}
function handleElementVNode(vnode, a) {
    var shouldSanitize = false;
    var attrs;
    var i;
    if (a) {
        for (i in a) {
            if ((shouldSanitize = CAMEL_PROPS.test(i))) {
                break;
            }
        }
        if (shouldSanitize) {
            attrs = vnode.attributes = {};
            for (i in a) {
                if (a.hasOwnProperty(i)) {
                    attrs[CAMEL_PROPS.test(i) ? i.replace(/([A-Z0-9])/, "-$1").toLowerCase() : i] = a[i];
                }
            }
        }
    }
}
// proxy render() since React returns a Component reference.
function render$1(vnode, parent, callback) {
    var prev = parent && parent._zreactCompatRendered && parent._zreactCompatRendered.vdom && parent._zreactCompatRendered.vdom.base;
    // ignore impossible previous renders
    if (prev && prev.parentNode !== parent) {
        prev = null;
    }
    // default to first Element child
    if (!prev && parent) {
        prev = parent.firstElementChild;
    }
    // remove unaffected siblings
    for (var i = parent.childNodes.length; i--;) {
        if (parent.childNodes[i] !== prev) {
            parent.removeChild(parent.childNodes[i]);
        }
    }
    var out = render(vnode, parent, prev);
    var vdom = findVDom(out);
    if (parent) {
        parent._zreactCompatRendered = (vdom && vdom.component) || { base: out };
    }
    if (typeof callback === "function") {
        callback();
    }
    return vdom && vdom.component || out;
}
var ContextProvider = /** @class */ (function () {
    function ContextProvider() {
    }
    ContextProvider.prototype.getChildContext = function () {
        return this.props.context;
    };
    ContextProvider.prototype.render = function (props) {
        return props.children[0];
    };
    return ContextProvider;
}());
function renderSubtreeIntoContainer(parentComponent, vnode, container, callback) {
    var wrap = h(ContextProvider, { context: parentComponent.context }, vnode);
    var renderContainer = render$1(wrap, container);
    var vdom = findVDom(renderContainer);
    var component = renderContainer._component || (vdom && vdom.base);
    if (callback) {
        callback.call(component, renderContainer);
    }
    return component;
}
function unmountComponentAtNode(container) {
    var existing = container._zreactCompatRendered && container._zreactCompatRendered.vdom && container._zreactCompatRendered.vdom.base;
    if (existing && existing.parentNode === container) {
        render(h(EmptyComponent, null), container, existing);
        return true;
    }
    return false;
}
/** Track current render() component for ref assignment */
var currentComponent;
function createFactory(type) {
    return createElement.bind(null, type);
}
var DOM = {};
for (var i = ELEMENTS.length; i--;) {
    DOM[ELEMENTS[i]] = createFactory(ELEMENTS[i]);
}
function upgradeToVNodes(arr, offset) {
    for (var i = offset || 0; i < arr.length; i++) {
        var obj = arr[i];
        if (Array.isArray(obj)) {
            upgradeToVNodes(obj);
        }
        else if (obj && typeof obj === "object" && !isValidElement(obj) && ((obj.props && obj.type) || (obj.attributes && obj.nodeName) || obj.children)) {
            arr[i] = createElement(obj.type || obj.nodeName, obj.props || obj.attributes, obj.children);
        }
    }
}
function isStatelessComponent(c) {
    return typeof c === "function" && !(c.prototype && c.prototype.render);
}
// wraps stateless functional components in a PropTypes validator
function wrapStatelessComponent(WrappedComponent) {
    return createClass({
        displayName: WrappedComponent.displayName || WrappedComponent.name,
        render: function () {
            return WrappedComponent(this.props, this.context);
        },
    });
}
function statelessComponentHook(Ctor) {
    var Wrapped = Ctor[COMPONENT_WRAPPER_KEY];
    if (Wrapped) {
        return Wrapped === true ? Ctor : Wrapped;
    }
    Wrapped = wrapStatelessComponent(Ctor);
    Object.defineProperty(Wrapped, COMPONENT_WRAPPER_KEY, { configurable: true, value: true });
    Wrapped.displayName = Ctor.displayName;
    Wrapped.propTypes = Ctor.propTypes;
    Wrapped.defaultProps = Ctor.defaultProps;
    Object.defineProperty(Ctor, COMPONENT_WRAPPER_KEY, { configurable: true, value: Wrapped });
    return Wrapped;
}
function createElement() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    upgradeToVNodes(args, 2);
    return normalizeVNode(h.apply(void 0, args));
}
function normalizeVNode(vnode) {
    vnode.zreactCompatNormalized = true;
    applyClassName(vnode);
    if (isStatelessComponent(vnode.nodeName)) {
        vnode.nodeName = statelessComponentHook(vnode.nodeName);
    }
    var ref = vnode.attributes.ref;
    var type = ref && typeof ref;
    if (currentComponent && (type === "string" || type === "number")) {
        vnode.attributes.ref = createStringRefProxy(ref, currentComponent);
    }
    applyEventNormalization(vnode);
    return vnode;
}
function cloneElement$1(element, props) {
    var children = [];
    for (var _i = 2; _i < arguments.length; _i++) {
        children[_i - 2] = arguments[_i];
    }
    if (!isValidElement(element)) {
        return element;
    }
    var elementProps = element.attributes || element.props;
    var node = h(element.nodeName || element.type, elementProps, element.children || elementProps && elementProps.children);
    // Only provide the 3rd argument if needed.
    // Arguments 3+ overwrite element.children in preactCloneElement
    var cloneArgs = [node, props];
    if (children && children.length) {
        cloneArgs.push(children);
    }
    else if (props && props.children) {
        cloneArgs.push(props.children);
    }
    return normalizeVNode(cloneElement.apply(void 0, cloneArgs));
}
function createStringRefProxy(name, component) {
    return component._refProxies[name] || (component._refProxies[name] = function (resolved) {
        if (component && component.refs) {
            component.refs[name] = resolved;
            if (resolved === null) {
                delete component._refProxies[name];
                component = null;
            }
        }
    });
}
function applyEventNormalization(_a) {
    var nodeName = _a.nodeName, attributes = _a.attributes;
    if (!attributes || typeof nodeName !== "string") {
        return;
    }
    var props = {};
    for (var i in attributes) {
        props[i.toLowerCase()] = i;
    }
    if (props.ondoubleclick) {
        attributes.ondblclick = attributes[props.ondoubleclick];
        delete attributes[props.ondoubleclick];
    }
    // for *textual inputs* (incl textarea), normalize `onChange` -> `onInput`:
    if (props.onchange && (nodeName === "textarea" || (nodeName.toLowerCase() === "input" && !/^fil|che|rad/i.test(attributes.type)))) {
        var normalized = props.oninput || "oninput";
        if (!attributes[normalized]) {
            attributes[normalized] = multihook([attributes[normalized], attributes[props.onchange]]);
            delete attributes[props.onchange];
        }
    }
}
function applyClassName(vnode) {
    var a = vnode.attributes || (vnode.attributes = {});
    classNameDescriptor.enumerable = "className" in a;
    if (a.className) {
        a.class = a.className;
    }
    Object.defineProperty(a, "className", classNameDescriptor);
}
var classNameDescriptor = {
    configurable: true,
    get: function () { return this.class; },
    set: function (v) { this.class = v; },
};
function extend(base, props) {
    for (var i = 1, obj = void 0; i < arguments.length; i++) {
        if ((obj = arguments[i])) {
            for (var key in obj) {
                if (obj.hasOwnProperty(key)) {
                    base[key] = obj[key];
                }
            }
        }
    }
    return base;
}
function shallowDiffers(a, b) {
    for (var i in a) {
        if (!(i in b)) {
            return true;
        }
    }
    for (var i in b) {
        if (a[i] !== b[i]) {
            return true;
        }
    }
    return false;
}
function F() { }
function createClass(obj) {
    function cl(props, context) {
        bindAll(this);
        Component$1.call(this, props, context, BYPASS_HOOK);
        newComponentHook.call(this, props, context);
    }
    obj = extend({ constructor: cl }, obj);
    // We need to apply mixins here so that getDefaultProps is correctly mixed
    if (obj.mixins) {
        applyMixins(obj, collateMixins(obj.mixins));
    }
    if (obj.statics) {
        extend(cl, obj.statics);
    }
    if (obj.propTypes) {
        cl.propTypes = obj.propTypes;
    }
    if (obj.defaultProps) {
        cl.defaultProps = obj.defaultProps;
    }
    if (obj.getDefaultProps) {
        cl.defaultProps = obj.getDefaultProps();
    }
    F.prototype = Component$1.prototype;
    cl.prototype = extend(new F(), obj);
    cl.displayName = obj.displayName || "Component";
    return cl;
}
// Flatten an Array of mixins to a map of method name to mixin implementations
function collateMixins(mixins) {
    var keyed = {};
    for (var _i = 0, mixins_1 = mixins; _i < mixins_1.length; _i++) {
        var mixin = mixins_1[_i];
        for (var key in mixin) {
            if (mixin.hasOwnProperty(key) && typeof mixin[key] === "function") {
                (keyed[key] || (keyed[key] = [])).push(mixin[key]);
            }
        }
    }
    return keyed;
}
// apply a mapping of Arrays of mixin methods to a component prototype
function applyMixins(proto, mixins) {
    for (var key in mixins) {
        if (mixins.hasOwnProperty(key)) {
            proto[key] = multihook(mixins[key].concat(proto[key]), key === "getDefaultProps" || key === "getInitialState" || key === "getChildContext");
        }
    }
}
function bindAll(ctx) {
    for (var i in ctx) {
        var v = ctx[i];
        if (typeof v === "function" && !v.__bound && !AUTOBIND_BLACKLIST.hasOwnProperty(i)) {
            (ctx[i] = v.bind(ctx)).__bound = true;
        }
    }
}
function callMethod(ctx, m, args) {
    if (typeof m === "string") {
        m = ctx.constructor.prototype[m];
    }
    if (typeof m === "function") {
        return m.apply(ctx, args);
    }
}
function multihook(hooks, skipDuplicates) {
    return function () {
        var ret;
        for (var _i = 0, hooks_1 = hooks; _i < hooks_1.length; _i++) {
            var hook = hooks_1[_i];
            var r = callMethod(this, hook, arguments);
            if (skipDuplicates && r != null) {
                if (!ret) {
                    ret = {};
                }
                for (var key in r) {
                    if (r.hasOwnProperty(key)) {
                        ret[key] = r[key];
                    }
                }
            }
            else if (typeof r !== "undefined") {
                ret = r;
            }
        }
        return ret;
    };
}
function newComponentHook(props, context) {
    propsHook.call(this, props, context);
    this.componentWillReceiveProps = multihook([propsHook, this.componentWillReceiveProps || "componentWillReceiveProps"]);
    this.render = multihook([propsHook, beforeRender, this.render || "render", afterRender]);
}
function propsHook(props, context) {
    if (!props) {
        return;
    }
    // React annoyingly special-cases single children, and some react components are ridiculously strict about this.
    var c = props.children;
    if (c && Array.isArray(c) && c.length === 1 && (typeof c[0] === "string" || typeof c[0] === "function" || c[0] instanceof VNode)) {
        props.children = c[0];
        // but its totally still going to be an Array.
        if (props.children && typeof props.children === "object") {
            props.children.length = 1;
            props.children[0] = props.children;
        }
    }
    // add proptype checking
    if (DEV) {
        var ctor = typeof this === "function" ? this : this.constructor;
        var propTypes = this.propTypes || ctor.propTypes;
        var displayName = this.displayName || ctor.name;
        if (propTypes) {
            PropTypes.checkPropTypes(propTypes, props, "prop", displayName);
        }
    }
}
function beforeRender(props) {
    currentComponent = this;
}
function afterRender() {
    if (currentComponent === this) {
        currentComponent = null;
    }
}
function Component$1(props, context, opts) {
    Component.call(this, props, context);
    this.state = this.getInitialState ? this.getInitialState() : {};
    this.refs = {};
    this._refProxies = {};
    if (opts !== BYPASS_HOOK) {
        newComponentHook.call(this, props, context);
    }
}
extend(Component$1.prototype = new Component({}, {}), {
    constructor: Component$1,
    isReactComponent: {},
    replaceState: function (state, callback) {
        this.setState(state, callback);
        for (var i in this.state) {
            if (!(i in state)) {
                delete this.state[i];
            }
        }
    },
    getDOMNode: function () {
        return this.vdom && this.vdom.base;
    },
    isMounted: function () {
        return !!(this.vdom && this.vdom.base);
    },
});
function PureComponent(props, context) {
    Component$1.call(this, props, context);
}
F.prototype = Component$1.prototype;
PureComponent.prototype = new F();
PureComponent.prototype.isPureReactComponent = true;
PureComponent.prototype.shouldComponentUpdate = function (props, state) {
    return shallowDiffers(this.props, props) || shallowDiffers(this.state, state);
};
var compat = {
    REACT_ELEMENT_TYPE: REACT_ELEMENT_TYPE,
    version: version,
    DOM: DOM,
    PropTypes: PropTypes,
    Children: Children,
    render: render$1,
    createClass: createClass,
    createFactory: createFactory,
    createElement: createElement,
    cloneElement: cloneElement$1,
    isValidElement: isValidElement,
    findDOMNode: findDOMNode,
    unmountComponentAtNode: unmountComponentAtNode,
    Component: Component$1,
    PureComponent: PureComponent,
    unstable_renderSubtreeIntoContainer: renderSubtreeIntoContainer,
    __spread: extend,
};

export { REACT_ELEMENT_TYPE, version, DOM, PropTypes, Children, render$1 as render, createClass, createFactory, createElement, cloneElement$1 as cloneElement, isValidElement, findDOMNode, unmountComponentAtNode, Component$1 as Component, PureComponent, renderSubtreeIntoContainer as unstable_renderSubtreeIntoContainer, extend as __spread };
export default compat;
//# sourceMappingURL=compat.js.map
